//	this file is the entry point for the application. 
//	The Gruntfile loads start.js, then all js files in src/javascript/, then end.js


// 'app' is the top level; all views, models, routers, etc are contained in 'app'.
var app = app || {};

// your public api url. Prod and Dev use this read-only, so it is staticly set here.
app.mediciUrl = "//" + window.location.host.replace(":8080", ":9001") + "/medici";

// loads jQuery and Backbone
(function($, Backbone, window, undefined) {
    "use strict";

var Sensors = Backbone.Collection.extend({
	url: app.mediciUrl + "/api/geostreams/sensors"
});

app.sensors = new Sensors();

// currently using this in the AdvancedSearchView.js, but may further explore collections with other models
var Datapoints = Backbone.Collection.extend({
	url: app.mediciUrl + "/api/geostreams/datapoints"
});

app.datapoints = new Datapoints();
// attach the event object to the app to ensure the object exists during the build process
app.event = {};
app.event.EventBus = _({}).extend(Backbone.Events);
// attach the model object to the app to ensure the object exists during the build process
app.model = {};
// the GeoStreamService model interfaces with '/datapoints' on the API

// if we have already collected the datapoints from the API, the 'geostream' argument will not be empty, and subsequent page loads will not hit the API. 
app.model.GeoStreamService = function(geostream, options) {
    this.hypercube = {};
    this.geostream = geostream ? geostream : {};
    this.options = {};
    // set the API endpoint for this model
    this.options.url = app.mediciUrl + "/api/geostreams/datapoints?";
    
    // 
    function populateGeostream(fn) {
        return function() {
            var args = arguments;
            var data = this.get(arguments[0].properties.name);
            if(data === null || data.length === 0) {
                $.getJSON(this.options.url, (function(self) {
                    return function(geostream) {
                        self.set(geostream);
                        fn.apply(self, args);
                    };
                })(this));
            } else {
                fn.apply(this, args);
            }
        };
    }
    
    // lookup a datapoint by location
    this.getByLocation = function(location, callback) {
        // if we have this location in our cached results, return it
        if(this.geostream[location.properties.name]) {
            callback(this.geostream[location.properties.name]);
            return;
        } else {
        // this location is not in our cached results, so fetch from the API using coordinates as query params
        var query = this.options.url + $.param({geocode: location.geometry.coordinates[1] + "," + location.geometry.coordinates[0] + "," + 1 });
        $.ajax(query, {
            dataType: "jsonp",
            success: (function(geostream) {
                return function(data) {
                    var ret = aggregateTime(location, data);
                
                    geostream[location.properties.name] = ret;
                    callback(ret);
                };
            })(this.geostream)
        });
        
        }
    };
    
    // getBy allows fetching based on a time frame
    this.getBy = function(location, startTime, endTime, parameters, callback) {
      // request from the API if this location isn't in our cached results
      // this will include the history of all records for this location
      if(!this.geostream[location.properties.name]) {
        fetchData(location, callback);
      }
      
      // if a start and end time are specified, we'll narrow the history to those times
      if(startTime && endTime) {
        var values = this.geostream[location.properties.name].sliceDate(Date.parse(startTime)/1000, Date.parse(endTime)/1000);
        values.
        callback();
      }
    };
    
    // fetchData retrieves data from the API for a specific location
    this.fetchData = function(location, callback) {
      var query = this.options.url + $.param({geocode: location.geometry.coordinates[1] + "," + location.geometry.coordinates[0] + "," + 1});
      var request = {
          dataType: "jsonp",
          success: (function(self) {
            return function(data) {
              self.storeData(location, data, callback);
            };
          })(this)
      };
      
      $.ajax(query, request);
    };
    
    this.storeData = function(location, data, callback) {
      var cubeData = data.map(function(d) {
        var values = {};
        location.properties.parameters.forEach(function(param) {
          if(isNaN(d.properties[param])) return;
          
          values[param] = +d.properties[param];
        });
        
        return {
          time: Date.parse(d.start_time.replace(" ", "T") + ":00") / 1000,
          facts: {
            station: location.properties.name,
            depth: d.geometry.coordinates[2]
          },
          measures: values
        };
      });
      
      ps.Cube.transforms.dateLocal(cubeData);
      
      return ps.Cube.deserialize(cubeData, location.parameters.map(function(p) { return p.replace(",", ""); }));
    };
    
    this.get = function(since, until, geocode) {
        return this.geostream;
    };
     
    this.set = function(geostream) {
        this.geostream = geostream ? geostream : [];
    };
};

function aggregateTime(location, datums, resolution) {
    var quantumMap = {};
    datums.forEach(function(datum) {
        var start_time = day(new Date(datum.start_time.replace(" ", "T") + ":00"));
        
        if(!quantumMap[start_time]) quantumMap[start_time] = [];
        
        quantumMap[start_time].push(datum);
        
        location.properties.parameters.forEach(function(d) {
            var prop = +datum.properties[d];
            if(!isNaN(prop)) {
                datum.properties[d.replace(",", "")] = prop;
            }
            
            delete datum.properties[d];
        });
    });
    
    return Object.keys(quantumMap).map(function(time) {
        var datetime = new Date(time);
        var timeString = (datetime.getUTCMonth() + 1) + "/" + datetime.getUTCDate() + "/" + datetime.getUTCFullYear();
        return { date: datetime, timestamp: timeString, values: aggregateDepth(location, quantumMap[time], 1) };
    });
    
    function day(date) {
        date.setUTCHours(0, 0, 0, 0);
        return date;
    }
    
    function month(date) {
        day(date).setUTCDate(0);
        return date;
    }
    
    function year(date) {
        month(date).setUTCMonth(0);
        return date;
    }
}

function aggregateDepth(location, datums, resolution) {
    var depthMap = {};
    datums.forEach(function(datum) {
        var bin = Math.floor(Math.abs(datum.geometry.coordinates[2]) / resolution);
        
        if(!depthMap[bin]) depthMap[bin] = [];
        
        depthMap[bin].push(datum);
    });
    
    return Object.keys(depthMap).map(function(bin) {
        var sum = depthMap[bin].reduce(function(a, b) {
            Object.keys(location.properties.parameters).forEach(function(key) {
                a.properties[key] += b.properties[key];
            });
            return a;
        });
        
        var count = depthMap[bin].length;
        
        Object.keys(sum.properties).forEach(function(key) {
            if(isNaN(sum.properties[key])) {
                delete sum.properties[key];
                return;
            }
            sum.properties[key] /= count;
        });
        
        return {depth: bin, data: sum.properties };
    });
}
// the StationService model interfaces with the '/stations' API endpoint

// like GeoStreamService, you can pass the argument 'stations' to the model on creation to prevent duplicate API requests
app.model.StationService = function(stations, options) {
    this.stations = stations ? this.set(stations) : [];
    this.options = options ? options : {};
    this.options.url = app.mediciUrl + "/api/geostreams/sensors";

    this.setStations = function(stations) {
      this.stations.length = 0;
      this.stations = stations;
    };
    
    this.getStations = function() {
      return this.stations;
    };
    
    this.fetchStations = function(callback) {
      if(this.getStations().length !== 0) {
        callback(this.getStations());
        return;
      }
      
      var request = {
          dataType: "jsonp"
      };
      
      request.success = (function(self) {
        return function(data) {
          self.setStations(data);
          callback(self.getStations());
        };
      })(this);
      
      $.ajax(this.options.url, request);
    };
    
    // the 'populateStationStats' function provides an interface to the API endpoint '/station/{id}/stats'
    this.populateStationStats = function(station, callback) {
      $.ajax(this.options.url + "/" + station.id + "/stats", {
        dataType: "jsonp",
        success: (function(self) {
          return function(data) {
            var updated = Object.keys(data.range).map(function(d) {
              if(!data.range[d]) return;
              var date = new Date(data.range[d].replace(" ", "T") + ":00");
              return (date.getUTCMonth() + 1) + "/" + date.getUTCDate() + "/" + date.getUTCFullYear();
            });
            
            station.properties.range = updated;
            station.properties.parameters = data.parameters;

            callback(station);
          };
        })(this)
      });
    };
    
    // getStationsBySources filters the existing 'stations' object (which is all stations) for all the unique "types" properties. "Sources" would typically be the organization that provides the data (EPA, USGS, etc.).  
    this.getStationsBySources = function(sources, callback) {
      this.fetchStations(function(stations) {
        var found = stations.filter(function(station) {
          return sources.indexOf(station.properties.type) >= 0;
        });
        
        callback(found);
      });
    };

    this.getAll = function(callback) {
      this.fetchStations(callback);
    };
    
    // getStationByName filters the 'stations' object for stations with the specified "name" property.
    this.getStationByName = function(location, callback) {
      this.fetchStations((function(self) {
        return function(stations) {
          var found = stations.filter(function(station) {
            return station.properties.name === location;
          });
        
          if(found[0].properties.parameters) {
            callback(found[0]); //assumes location names are unique
          } else {
            self.populateStationStats(found[0], callback);
          }
        };
      })(this));
    };
};
app.Sensor = Backbone.Model.extend({
	urlRoot: app.mediciUrl + "/api/geostreams/sensors",
    initialize: function(){
    }
});

app.router = {};
app.router.AppRouter = Backbone.Router.extend({
    
    routes : {
        "" : "home",
        "explore" : "home",
        "search" : "advancedSearch",
        "detail/location/:location/" : "populateTimes",
        "detail/location/:location/startTime/:startTime/endTime/:endTime/" : "populateParameters",
        "detail/location/:location/startTime/:startTime/endTime/:endTime/parameters/:parameters/" : "populateVisualization",
        "detail/location/:location/startTime/:startTime/endTime/:endTime/parameters/:parameters/depth/:depth" : "populateDepthVisualization",
    }
});
app.view = {};
app.view.AppView = Backbone.View.extend({
    el : "#app",
    
    events : {
      "click #header-go-home" : "goHome"
    },
    
    initialize : function() {
        this.router = new app.router.AppRouter();
        
        this.eventBus = _({}).extend(Backbone.Events);
        
        this.marker = _.template($('#marker-template').html())();

        this.stationService = new app.model.StationService();
        this.geoStreamService = new app.model.GeoStreamService();
        
        this.currentView = null;
        
        this.homeView = new app.view.HomeView({
          parentView: this,
          eventBus: this.eventBus,
          marker: this.marker,
          router: this.router
        });
        
        $(window).resize((function(self) {
            return function(event) {
                self.render();
            };
        })(this));
        
        this.router.on("route:home", (function(self) {
            return function() {
                //$("#map-nav").css("display", "block");
                //$("#detail-panel").css("display", "none");
                $(".navbar > li").removeClass("active");
                $("a[href='#explore']").parent().addClass("active");
                //$("#advanced-search-panel").css("display", "none");
                
                self.setView(self.homeView)
            };
        })(this));
        
        this.router.on("route:populateTimes", (function(self) {
            return function(location) {
                self.initializeDetailView(location);
            };
        })(this));
        
        this.router.on("route:populateParameters", (function(self) {
            return function(location, startTime, endTime) {
                self.initializeDetailView(location, startTime, endTime);
            };
        })(this));
        
        this.router.on("route:populateVisualization", (function(self) {
            return function(location, startTime, endTime, parameters) {
                self.initializeDetailView(location, startTime, endTime, parameters);
            };
        })(this));
        
        this.router.on("route:populateDepthVisualization", (function(self) {
            return function(location, startTime, endTime, parameter, depth) {
                self.initializeDetailView(location, startTime, endTime, parameter, depth);
            };
        })(this));
        
        this.router.on("route:advancedSearch", (function(self) {
            return function() {
                self.initializeAdvancedSearchView();
            };
        })(this));
        
        this.eventBus.bind("change:location", (function(self) {
            return function(location) {
                var loc = "location/" + location + "/";
                self.router.navigate("detail/" + loc, {trigger : true, replace : true});
            };
        })(this));
        
        this.eventBus.bind("change:depth", (function(self) {
            return function(depth) {
                depth = "depth/" + depth;
                if(Backbone.history.fragment.indexOf("depth") < 0) {
                    self.router.navigate(Backbone.history.fragment + depth, {trigger : true, replace : true});
                } else {
                    var newTarget = Backbone.history.fragment.replace(/\/depth\/[^\/]*\/?/, "/" + depth);
                    self.router.navigate(newTarget, {trigger : true, replace : true});
                }
            };
        })(this));
        
        this.eventBus.bind("change:parameters", (function(self) {
            return function(parameters) {
                var par =  parameters.length === 0 ? "" : "parameters/" + parameters.join() + "/";
        
                if(Backbone.history.fragment.indexOf("parameters") < 0) {
                    self.router.navigate(Backbone.history.fragment + par, {trigger : true, replace : true});
                } else {
                    var replace = parameters.length === 0 ? /parameters\/[^\/]*\/(?:depth\/[^\/])?/ : /parameters\/[^\/]*\//;
                    var newTarget = Backbone.history.fragment.replace(replace, par);
                    self.router.navigate(newTarget, {trigger : true, replace : true});
                }
            };
        })(this));
        
        this.eventBus.bind("change:times", (function(self) {
            return function(start, end) {
                var t = "startTime/" + encodeURIComponent(start) + "/endTime/" + encodeURIComponent(end) + "/";
        
                if(Backbone.history.fragment.indexOf("startTime") < 0) {
                    self.router.navigate(Backbone.history.fragment + t, {trigger : true, replace : true});
                } else {
                    var newTarget = Backbone.history.fragment.replace(/startTime\/[^\/]*\/endTime\/[^\/]*\//, t);
                    self.router.navigate(newTarget, {trigger : true, replace : true});
                }
            };
        })(this));
        
        this.eventBus.bind("change:sources", (function(self) {
            return function(sources) {
                self.stationService.getStationsBySources(sources, function(data) {
                    self.mainMap.setGeoJsonFeature(data);
                    self.mainMap.render();
                });
            };
        })(this));
        
        this.eventBus.bind("station:fetch", (function(self) {
          return function(stationName) {
            
            self.stationService.getStationByName(stationName, function(station) {
              self.eventBus.trigger("station:update", station);
            });
          }
        })(this));
        
        this.stationService.getAll((function(self) {
          return function(data) {
            self.homeView.setData(data);
          };
          
        })(this));
        
        Backbone.history.start();
    },
    
    render : function() {
        if(this.currentView) this.currentView.render();
    },
    
    goHome : function() {
      this.router.navigate("", {trigger : true, replace : true});
    },
    
    setView : function(view) {
      if(this.currentView !== view) {
        this.currentView = view;
        this.$("#content-container").empty()
          .append(this.currentView.el);
      }
      
      this.currentView.render();
    },
    
    initializeDetailView : function() {
      if(!this.detailView) {
        this.detailView = new app.view.DetailView({
          parentView : this,
          eventBus : this.eventBus,
          marker : this.marker,
          router : this.router,
          stationService : this.stationService,
          geostreamService : this.geoStreamService
        });
      }
      
      this.detailView.initiate.apply(this.detailView, arguments);
      
      this.setView(this.detailView);
    },
    
    initializeAdvancedSearchView : function() {
      if(!this.advancedSearchView) {
        this.advancedSearchView = new app.view.AdvancedSearchView({
          parentView : this,
          eventBus : this.eventBus,
          router : this.router,
          stationService : this.stationService,
          geostreamService : this.geoStreamService
        });
      }
      
      this.setView(this.advancedSearchView);
    }
});
app.view.ControlsView = Backbone.View.extend({
	tagName : 'div',
	id : 'controls',

	events : {
		'change #times' : 'changeTime',
		'change input[name="parameter"]' : 'changeParameters',
	},
	
	initialize : function(options) {
	  this.eventBus = options.eventBus;
		this.render = _.bind(this.render, this);
	},
	
	render : function() {
		this.renderParameters(this.parameters)
			.renderSlider(this.times);
	},
	
	renderSlider : function(items, location) {
		var $slider = d3.select("#slider");
		
		var data = { items: items, location: location };
        
        var instance = $slider.selectAll(".instance").data([data], function(d) { return d.location; });
        instance.enter().append("div").attr("class", "instance")
        .call((function (self) {
            return function(selection) {
                selection.each(function(d) {
                    $(this).slider({
                        range : true,
                        step : 1,
                        values : [0, d.items.length - 1],
                        min : 0,
                        max : d.items.length - 1,
                        change : function(event, ui) {
                            var values = ui.values;
                            self.changeTimes(values[0], values[1]);
                        },
                        create : function(event, ui) {
                            self.changeTimes(0, d.items.length - 1);
                        },
                        slide: function(event, ui) {
                            var values = ui.values;
                            self.sliderTimes(self.times[values[0]], self.times[values[1]]);
                        }
                    });
                });
            };
        })(this));

        instance.exit().remove();
		
		return this;
	},
	
	renderParameters : function(parameters, selected, location) {
        var $parameters = d3.selectAll("#parameters");
        
        var data = parameters.map(function(d) {
            return { parameter: d, location: location };
        });
        
        var controlGroup = $parameters.selectAll(".control-group").data(data, function(d) { return d.location + "." + d.parameter; });
        controlGroup.enter().append("div").attr("class", "control-group");
        controlGroup.exit().remove();
        
        var controlLabel = controlGroup.selectAll(".control-label").data(function(d) { return [d]; });
        controlLabel.enter().append("label");
        controlLabel.attr("class", function(d) { return "control-label " + d.parameter; })
            .attr("for", function(d) { return d.parameter; }).html(function(d) { return d.parameter.replace(/Chlorophyll( |-)a/, "Chlorophyll <i>a</i>"); });
            
        var controls = controlGroup.selectAll(".controls").data(function(d) { return [d]; });
        controls.enter().append("div").attr("class", "controls");
        
        var input = controls.selectAll("input").data(function(d) { return [d]; });
        input.enter().append("input").attr("type", "checkbox").attr("name", "parameter")
            .property("checked", function(d, i) { if(selected) return selected.filter(function(e) { return e === d.parameter; } ).length === 0 ? null : true; })
            .attr("id", function(d) { return d.parameter; }).attr("value", function(d) { return d.parameter; })
            .on("click", function(self) {
              self.changeParameters();
                return function(event) {
                    self.changeParameters();
                };
            }(this));
		
		return this;
	},
    
    setParameters : function(parameters, selected) {
        this.parameters = parameters;
        this.renderParameters(parameters, selected, this.location());
    },
    
    setTimes : function(times) {
        this.times = times;
        this.renderSlider(times, this.location());
    },
    
    location : function(location) {
        if(!location) return this.loc;
        
        this.loc = location;
    },
    
    sliderTimes : function(start, end) {
      var prettyStart = new Date(start);
      prettyStart = (prettyStart.getUTCMonth() + 1) + "/" + prettyStart.getUTCDate() + "/" + prettyStart.getUTCFullYear();
    
      var prettyEnd = new Date(end);
      prettyEnd = (prettyEnd.getUTCMonth() + 1) + "/" + prettyEnd.getUTCDate() + "/" + prettyEnd.getUTCFullYear();
    
      $('#date-range').html(prettyStart + ' - ' + prettyEnd);
    },
    
    changeTimes : function(start, end) {
      start = this.times[start];
      end = this.times[end];
		
      this.sliderTimes(start, end);
        
      this.eventBus.trigger("change:times", start, end);	
    },
	
	changeParameters : function() {
		var vals = [];
		$('#parameters input[name="parameter"]:checked').each(function() {
			vals.push(encodeURIComponent($(this).attr("value")));
		});
        
        this.eventBus.trigger("change:parameters", vals);
	},
    
    getSelectedParameters : function() {
        var ret = [];
        
        $('#parameters input[type="checkbox"]:checked').each(function() {
            ret.push($(this).val());
        });
        
        return ret;
    }
});
app.view.DetailView = Backbone.View.extend({
  id : "detail-view",
  tagName : 'div',
	className : 'row-fluid',
    
  events : {
    "click #detail-go-back" : "goBack"
  },
	
	initialize : function(options) {
	  
	  this.stationService = options.stationService;
	  this.geostreamService = options.geostreamService;
	  this.router = options.router;
	  this.eventBus = options.eventBus;
	  this.marker = options.marker;
	  
	  this.$el.append(_.template($("#detail-panel-template").html())());
        
		this.miniMap = new app.view.MiniMapView({
		  tagName : 'div',
      parentView : this,
      marker : this.marker,
      mapOptions : {
        zoomControl : false,
        attributionControl : false,
        dragging : false,
        touchZoom : false,
        scrollWheelZoom : false,
        doubleClickZoom : false,
        boxZoom : false
      },
      geoJsonFeature : []
    });
		
		this.$("#mini-map").append(this.miniMap.el);
		
		this.controlsView = new app.view.ControlsView({
		  el : "#modal-content",
      parentView : this,
      eventBus : this.eventBus
    });
        
    this.aggregateTimeSeriesView = new app.view.TimeSeriesView({
      parentView : this,
    });
    
    this.$("#aggregate-timeseries").append(this.aggregateTimeSeriesView.el);
		
		this.limnologyView = new app.view.LimnologyView({
      parentView : this,
    });
		
		this.$("#limnology-chart").append(this.limnologyView.el);
        
    this.limnologyView.changeDepth((function(self) {
      return function(i) {
        self.eventBus.trigger("change:depth", i);
      };
		})(this));
		
		this.timeSeriesView = new app.view.TimeSeriesView({
      el : "#timeseries-at-depth",
      parentView : this,
		});
        
		$('a[data-toggle="tab"]').on("shown", function(self) {
		  return function(event) {
		    self.render();
      };
    }(this));
        
    $("#showDistribution").on("click", (function(self) {
      return function() {
        $(this).button("toggle");
          self.aggregateTimeSeriesView.render();
          self.timeSeriesView.render();
      };
    })(this));
	},
    
    initiate : function(location, startTime, endTime, parameters, depth) {
        this.stationService.getStationByName(location, (function(self) {
            return function(station) {
                self.geostreamService.getByLocation(station, function(data) {
                    self.controlsView.location(location);
                    var times = data.map(function(d) { return d.date.toUTCString(); });
                    self.controlsView.setTimes(times);
                    
                    if(!startTime) return;
                    
                    var trackedParams = station.properties.parameters.map(function(f) { return f.replace(",", ""); });
                    
                    var p = (!parameters ? self.controlsView.getSelectedParameters() : parameters.split(",")).map(function(f) { return f.valueOf().trim(); });
                    
                    p = p.length ? p : trackedParams;
                    
                    if(startTime) self.controlsView.setParameters(trackedParams, p);
                    
                    var par = {};
                    data.forEach(function(d) {
                        var datetime = d.date;
                        var dateString = d.date.toUTCString();
                        if(!(datetime.getTime() >= (new Date(startTime)).getTime() && datetime.getTime() <= (new Date(endTime)).getTime())) return;
                        var sum = d.values.reduce(function(a, b) {
                            trackedParams.forEach(function(key) {
                                a.data[key] += b.data[key];
                            });
                            
                            return a;
                        });
                        
                        var count = d.values.length;
                        
                        Object.keys(sum.data).forEach(function(parameter) {
                            sum.data[parameter] /= count;

                            for(var i = 0; i < p.length; i++) {
                                if(parameter === p[i] && !isNaN(sum.data[parameter])) {
                                    if(!par[i]) par[i] = {};
                                    
                                    if(!par[i][parameter]) par[i][parameter] = {};
                                    
                                    if(!par[i][parameter][dateString]) par[i][parameter][dateString] = {};
                                    
                                    par[i][parameter][dateString] = { x : dateString, y: sum.data[parameter] };
                                }
                            }
                        });
                    });
                    
                    var display_data = Object.keys(par).map(function(e) {
                        return { key: e, values: Object.keys(par[e]).map(function(f) { return { key: f, values: Object.keys(par[e][f]).map(function(g) { return par[e][f][g]; })}; }) };
                    });
                    
                    self.aggregateTimeSeriesView.setData(display_data);
                    self.limnologyView.setData(self.getDepthRep(data, p).filter(function(d) { return (new Date(d.key).getTime()) === (new Date(startTime)).getTime(); }));
                    
                    $("#detail-view-header").html(location);
                
                    self.miniMap.setGeoJsonFeature([station]);
                    self.miniMap.render();
                
                    self.render();
                });
            };
        })(this));
    },
	
	render : function() {
        this.$("#time-series.tab-pane.active").each((function(self) { return function() { self.aggregateTimeSeriesView.render(); }; })(this));
        this.$("#depth-profile.tab-pane.active").each((function(self) { return function() { self.limnologyView.render(); self.timeSeriesView.render(); }; })(this));
        this.miniMap.render();
	},
    
    goBack : function(event) {
        this.router.navigate("", {trigger: true, replace: false});
    },
    
    getDepthRep: function(data, parameters) {
        return data.map(function(times) {
            var par = {};
            
            times.values.forEach(function(depth) {
                Object.keys(depth.data).forEach(function(parameter) {
                    var param = parameter.replace(",", "");
                    parameters.forEach(function(p) {
                        if(param !== p) return;
                        
                        if(!par[param]) par[param] = [];
                    
                        par[param].push({ x: depth.data[parameter], y: +depth.depth, x0: 0, x1: 0 });
                    });
                });
            });
            
            
            return {
                key: times.date,
                values: Object.keys(par).map(function(param) {
                    return { key: param, values: par[param] };
                })
            };
        });
    }
});
app.view.ExploreView = Backbone.View.extend({
   tagName : "div",
   className : "accordian",
   lockedMarker : null,
   
   initialize : function(options) {
       this.marker = options.marker;
       this.eventBus = options.eventBus;
       this.router = options.router;
       
       this.render = _.bind(this.render, this);
       
       this.template = _.template($('#sources-nav').html());
       
       this.modals = _.template($("#seagrant-modals").html())();
       
       this.infoIcon = _.template($('#info-icon').html())();
   },
   
   render : function() {
       this.$el.empty();
       
       this.sources.forEach((function(self) {
           return function(d, i) {
               self.$el.append(self.template({
                        id : d.id,
                        elementId : self.el.id,
                        svgMarker : self.marker,
                        infoIcon : self.infoIcon,
                        title : d.title,
                        index : i,
                        regions : d.regions,
                        source : d.source
                    }
                ));
           };
        })(this));
       
       this.$el.append(this.modals);
        
        var $explorePills = this.$('.explore-pill');
        $explorePills.tooltip({animation : true, placement : "right", container : "body"}); // enable bootstrap tooltips
        
        $explorePills.hover(function(self) {
            return function(event) {
                if(self.markerLock) return;
                
                self.eventBus.trigger("marker:mouseover", event.currentTarget.value);
            };
        }(this), function(self) {
            return function(event) {
                if(self.markerLock) return;
                
                self.eventBus.trigger("marker:mouseout", event.currentTarget.value);
            };
        }(this));
        
        $explorePills.on("click", (function(self) {
            return function(event) {
                if(self.markerLock && event.currentTarget.value === self.markerLock) {
                    self.markerLock = null;
                    self.eventBus.trigger("marker:unclick", event.currentTarget.value);
                } else {
                    self.markerLock = event.currentTarget.value;
                }
                self.eventBus.trigger("marker:click", event.currentTarget.value);
            };
        })(this));
        
        $explorePills.on("dblclick", (function(self) {
            return function(event) {
                self.eventBus.trigger("change:location", event.currentTarget.value);
            };
        })(this));
        
        /* Keep the checkboxes in sync with the collapsed accordion
         * checked - accordion open
         * unchecked - accordion closed
         * not happy with this solution.
         */
        var $collapse = this.$('.collapse');
        
        $collapse.on("hide", (function(self) {
            return function(event) {
                if(event.target !== this) return;
                $(this).parent().find('input[name="source"]').prop("checked", false);
                
                var sources = [];
                $('input[name="source"]:checked').each(function() {
                    sources.push($(this).attr("value"));
                });
                self.eventBus.trigger("change:sources", sources);
            };
        })(this));
        
        $collapse.on("show", (function(self) {
            return function(event) {
                if(event.target !== this) return;
                $(this).parent().find('input[name="source"]').prop("checked", true);
                
                var sources = [];
                $('input[name="source"]:checked').each(function() {
                    sources.push($(this).attr("value"));
                });
                self.eventBus.trigger("change:sources", sources);
            };
        })(this));
    },
    
    setSources : function(sources) {
        this.sources = sources;
    }
});
app.view.HomeView = Backbone.View.extend({
  id: 'home-view',
  tag: 'div',
  
  initialize: function(options) {
    this.marker = options.marker;
    this.eventBus = options.eventBus;
    this.router = options.router;
    
    this.render = _.bind(this.render, this);

    this.$el.append(_.template($("#explore-template").html())());
    
    this.exploreView = new app.view.ExploreView({
      parentView: this,
      eventBus : this.eventBus,
      marker : this.marker,
      router : this.router,
    });
    
    this.$("#exploreView").append(this.exploreView.el);
  
    this.mainMap = new app.view.MainMapView({
      parentView : this,
      eventBus : this.eventBus,
      marker : this.marker,
      router : this.router,
      geoJsonFeature : []
    });
    
    this.$("#data-view").append(this.mainMap.el);
  },
  

  setData : function(data) {
    var sources = {};
    data.forEach(function(d) {
      if (!sources[d.properties.type.id])
        sources[d.properties.type.id] = [];

      sources[d.properties.type.id].push(d);
    });

    var output = Object.keys(sources).map(function(source) {
    var regions = {};

    var title = sources[source][0].properties.type.title;

    sources[source].forEach(function(d) {
      if (!regions[d.properties.region])
        regions[d.properties.region] = [];

      regions[d.properties.region].push(d);
    });

      return {
        id : source,
        title : title,
        source : source.toUpperCase(),
        regions : Object.keys(regions).map(function(r) {
          return {
            id : r,
            stations : regions[r].map(function(f) {
              return f;
            })
          };
        })
      };
    });

    // TODO: move this into the new HomeView.js
    this.exploreView.setSources(output);
    this.mainMap.setGeoJsonFeature(data);
    this.render();
  },
  
  render: function() {
    this.exploreView.render();
    this.mainMap.render();
  }
});
app.view.LimnologyView = Backbone.View.extend({
	tagName : 'div',
	id : 'main-v11n',
	className : 'span4',
	
	initialize : function() {
		this.render = _.bind(this.render, this);
		
        this.data = [];
        
		this.dispatch = d3.dispatch("changeDepth");
		
		this.depthGraph = v11n.layout.depth()
			.height(400).width(309).dispatch(this.dispatch);
		
		this.$el.append('<h4 id="main-viz-title">Depth Profile</h4>');
	},
	
	render : function() {
        var container = d3.select(this.el);
		var graph = container.selectAll(".depth-chart").data(this.data),
		graphEnter = graph.enter().append("div").attr("class", "depth-chart"),
		graphExit = graph.exit().remove();
		graph.call(this.depthGraph);
	},
    
    setData : function(data) {
        this.data = data;
    },
    
    changeDepth : function(callback) {
        this.dispatch.on("changeDepth", callback);
    }
});
app.view.MapView = Backbone.View.extend({
      tagName : "div",
      className : "leaflet_map",

      initialize : function(options) {
        this.mapOptions = options.mapOptions;
        this.eventBus = options.eventBus;
        this.marker = options.marker;
        
        this.render = _.bind(this.render, this);

        this.map = L.map(this.el, this.mapOptions);

        this.tileLayer = L
            .tileLayer(
                'http://{s}.tile.osm.org/{z}/{x}/{y}.png',
                {
                  attribution : '&copy; <a href="http://osm.org/copyright">OpenStreeMap</a> contributors'
                });

        this.setGeoJsonFeature(options.geoJsonFeature);
      },

      render : function() {
        this.map.invalidateSize(true);
        this.tileLayer.addTo(this.map);
        this.geoJsonLayer.addTo(this.map);
      },

      setGeoJsonFeature : function(geoJsonFeature) {
        if (this.geoJsonLayer)
          this.map.removeLayer(this.geoJsonLayer);

        this.geoJsonFeature = geoJsonFeature;
        this.geoJsonLayer = L.geoJson(this.geoJsonFeature, this
            .geoJsonOptions(this));
      }
});
app.view.MainMapView = app.view.MapView.extend({
    tagName : "div",
    className : "leaflet_map",
    markerLock : null,
    
    markers : {},
    latlngFormat : d3.format(".4f"),
    
    geoJsonOptions : function(self) {
        return {
            pointToLayer: function(feature, latlng) {
                var marker = L.marker(latlng, {
                    icon : L.divIcon({
                        html : '<div id="' + ("station-" + feature.id) + '" class="' + feature.properties.type.id + '">' + self.marker + '</div>'
                    }),
                    title : feature.properties.name,
                    riseOnHover : true
                });
                
                self.markers["station-" + feature.properties.name] = marker;
                
                return marker;
            },
            onEachFeature: function(feature, layer) {
                if (feature.properties && feature.properties.popupContent) {
                    var popupContent = self.popupTemplate(
                        {
                            source: feature.properties.type.id,
                            stationId: feature.properties.name,
                            stationName: feature.properties.popupContent,
                            sourceType: feature.properties.type.title,
                            timePeriod: 'fetching...',
                            lat: self.latlngFormat(layer.getLatLng().lat),
                            lon: self.latlngFormat(layer.getLatLng().lng),
                            parameters: ['fetching...']
                        }
                    );

                    layer.bindPopup(popupContent, { offset: new L.Point(160, -50)});
                    layer.on("mouseover", function(event) {
                        if(event.originalEvent.target.className.baseVal === "mouseCapture") {
                            self.highlightMarker("station-" + feature.properties.name);
                        }
                    });
                    layer.on("mouseout", function(event) {
                        if(event.originalEvent.target.className.baseVal === "mouseCapture") {
                            self.unhighlightMarker("station-" + feature.properties.name);
                        }
                    });
                    layer.on("click", function(event) {
                        self.eventBus.trigger("change:location", feature.properties.name);
                        self.unhighlightMarker("station-" + feature.properties.name);
                    });
                }
            }
        };
    },
    
    initialize : function(options) {
        this.popupTemplate = _.template($("#info-window").html());
        
        app.view.MapView.prototype.initialize.apply(this, arguments);
        
        this.eventBus.bind("marker:mouseover", (function(self) {
            return function(markerId) {
                markerId = "station-" + markerId;
                self.highlightMarker(markerId);
            };
        })(this));
        
        this.eventBus.bind("marker:mouseout", (function(self) {
            return function(markerId) {
                markerId = "station-" + markerId;
                self.unhighlightMarker(markerId);
            };
        })(this));
        
        this.eventBus.bind("marker:click", (function(self) {
            return function(markerId) {
                markerId = "station-" + markerId;
                if(self.markerLock && self.markerLock !== markerId) {
                    self.unhighlightMarker(self.markerLock);
                }
                
                self.highlightMarker(markerId);
                self.markerLock = markerId;
                
                self.panToMarker(markerId);
            };
        })(this));
        
        this.eventBus.bind("marker:unclick", (function(self) {
            return function(markerId) {
                markerId = "station-" + markerId;
                self.unhighlightMarker(markerId);
                self.markerLock = null;
            };
        })(this));
        
        this.eventBus.bind("station:update", (function(self) {
          return function(station) {
            var popupContent = self.popupTemplate(
                {
                  source: station.properties.type.id,
                  stationId: station.properties.name,
                  stationName: station.properties.popupContent,
                  sourceType: station.properties.type.title,
                  timePeriod: station.properties.range[0] + "-" + station.properties.range[1],
                  lat: self.latlngFormat(station.geometry.coordinates[1]),
                  lon: self.latlngFormat(station.geometry.coordinates[0]),
                  parameters: station.properties.parameters
              }
            );
            
            self.markers["station-" + station.properties.name].setPopupContent(popupContent, { offset: new L.Point(160, -50)});
          };
        })(this));
    },
    
    render : function() {
        app.view.MapView.prototype.render.apply(this, arguments);
        this.map.fitBounds((new L.featureGroup([this.geoJsonLayer])).getBounds().pad(0.05));
    },
    
    setGeoJsonFeature : function(geoJsonFeature) {
        app.view.MapView.prototype.setGeoJsonFeature.call(this, geoJsonFeature);
        
        if(this.geoJsonFeature && this.geoJsonFeature.length) {
            this.render(); //this.map.fitBounds((new L.featureGroup([this.geoJsonLayer])).getBounds().pad(0.05));
        }
    },
    
    highlightMarker : function(markerId) {
        d3.selectAll('#' + markerId + ' svg').attr("class", "map-marker-hover");
        this.eventBus.trigger("station:fetch", this.markers[markerId].options.title);
        
        this.markers[markerId].openPopup();
    },
    
    unhighlightMarker : function(markerId) {
        d3.selectAll('#' + markerId + ' svg').attr("class", "map-marker");
        this.markers[markerId].closePopup();
    },
    
    panToMarker : function(markerId) {
        this.map.panTo(this.markers[markerId].getLatLng());
    }
});
app.view.MiniMapView = app.view.MapView.extend({
    geoJsonOptions : function(self) {
        return {
            pointToLayer : function(feature, latlng) {
                var marker = L.marker(latlng, {
                    icon : L.divIcon({
                        html : '<div id="' + feature.properties.name + '" class="' + feature.properties.type.id + '">' + self.marker + '</div>'
                    })
                });
                
                return marker;
            },
            onEachFeature : function(feature, layer) {
                
            }
        };
    },
    
    initialize : function(options) {
        app.view.MapView.prototype.initialize.apply(this, arguments);
    },
    
    render : function() {
        app.view.MapView.prototype.render.apply(this, arguments);
    },
    
    setGeoJsonFeature : function(geoJsonFeature) {
        if(geoJsonFeature.length === 0) return;
      
        app.view.MapView.prototype.setGeoJsonFeature.apply(this, arguments);
        
        var coordinates = geoJsonFeature[0].geometry.coordinates;
        this.map.setView(new L.LatLng(coordinates[1], coordinates[0]), 7);
    }
});
app.view.TimeSeriesView = Backbone.View.extend({
	tagName : 'div',
	id : 'time',
    
    defaultOptions : {
        height : 200,
        width : 525,
        boxWhiskerMarginH : 40,
        boxWhiskerMarginV : 40
    },
	
	initialize : function(options) {
		this.render = _.bind(this.render, this);
        
        this.data = [];
		
        this.chartOptions = _.extend({}, this.defaultOptions, options.chartOptions);
        
		this.timeSeries = v11n.layout.timeSeries().height(this.chartOptions.height).width(this.chartOptions.width).max(15).padding("10 10 40 60");
		this.boxWhisker = v11n.layout.boxwhisker().height(this.chartOptions.height).width(175).margin(this.chartOptions.boxWhiskerMarginH, this.chartOptions.boxWhiskerMarginV);
		this.$el.append('<h4 id="time-series-title">Time Series</h4>');
	},
	
	render : function() {
        if(!this.data || !this.data.length) {
            this.$el.html('<div class="no-data">No data for the selected attributes. Have you selected a parameter?</div>');
        } else {
            this.$('.no-data').remove();
            
            var container = d3.select(this.el).selectAll(".time-series-distribution-charts").data(this.data, function(d) { return d.key; });
            container.enter().append("div").attr("class", "time-series-distribution-charts row-fluid");
            container.exit().transition().remove();
            
            var showDistribution = $("#showDistribution").hasClass("active");
            
            var mytimeseries = container.selectAll(".time-series-chart").data(function(d) { return [d]; }, function(d) { return d.key; });
            mytimeseries.enter().append("div").attr("class", "time-series-chart");
            mytimeseries.exit().remove();
            mytimeseries.classed("span9", showDistribution).classed("span12", !showDistribution)
                .call(this.timeSeries);
            
            var boxWhiskerChart = container.selectAll(".box-whisker-chart").data(function(d) { return showDistribution ? [d] : []; }, function(d) { return d.key; });
            boxWhiskerChart.enter().append("div").attr("class", "box-whisker-chart span3");
            boxWhiskerChart.exit().remove();
            boxWhiskerChart.call(this.boxWhisker);
        }
	},
    
    setData : function(data) {
        this.data = data;
    }
});
app.view.AdvancedSearchView = Backbone.View.extend({
	tagName : 'div',
	id : "advanced-search-body",
	className : 'row-fluid',
    
    events : {
        "click #download-permalink" : "permalink"
    },
	
	initialize : function(options) {
        console.log("Initializing advanced search view");
        
        this.$el.append(_.template($("#advanced-search-template").html())());
        
        this.eventBus = options.eventBus;
        this.stationService = options.stationService;
        //this.template = _.template($('#advanced-search-panel').html());
        // container for currently selected sensors
        this.sensorsfound = L.layerGroup();
        this.sensors = {};
        this.coordinates = {};
        this.format = 'YYYY-MM-DD h:mm:ss';
        this.startDate = moment('2010-01-01 1:00:00');
        this.endDate = moment('2010-06-01 1:00:00');
        
        var that = this;
        
        // map
        this.map = L.map(this.$('#advanced-map')[0]);
        
        // openstreet
        var openstreet = L.tileLayer('http://{s}.tile.osm.org/{z}/{x}/{y}.png', {
            attribution: 'Map data &copy; <a href="http://osm.org/copyright">OpenStreeMap</a> contributors',
            maxZoom: 18
        });
        openstreet.addTo(this.map);

        // controls
        var baseMaps = {
            "OpenStreetMap": openstreet
        };

        this.mapControls = L.control.layers(baseMaps);
        this.mapControls.addTo(that.map);

        this.eventBus.on("sensors:selected", (function(self) {
            return function(sensors) {
                that.sensors = sensors;
                console.log("sensors:selected");
                console.log(that.sensors);
            };
        })(this));

        this.eventBus.on("map:circle:select", (function(self) {
            return function(coordinates) {
                that.coordinates = coordinates;
                console.log("map:circle:select");
                console.log(that.coordinates);
                that.updateDownloadLink();
            };
        })(this));


        // examples of how to query sensor collection

        // var sensor = new app.Sensor({id: 761});

        // sensor.fetch({
        //     dataType: "jsonp",
        //     success: function (sensor) {
        //         console.log('Sensor model by id ' + sensor.toJSON()['name']);
        //     }
        // })

        // var sensorFromCollection = app.sensors.get('761');
        // console.log('Sensor from collection ' + sensorFromCollection);

        // app.sensors.fetch({
        //     dataType: "jsonp",
        //     data: $.param({geocode: '46.77,-87.55,100.0'}),
        //     success: function (sensors) {
        //         console.log(sensors.toJSON());
        //     }
        // });

	},
	
	render : function() {
        var that = this;
        this.$el.css("display", "block");

        $(".navbar > li").removeClass("active");
        $("a[href='#search']").parent().addClass("active");
        $("#map-nav").css("display", "none");
        this.$el.html(this.template);

        // time range selector
        $('#searchtime').val(this.startDate.format(this.format) +  " - " + this.endDate.format(this.format))
        $('#searchtime').daterangepicker(
          { 
            format: 'YYYY-MM-DD h:mm:ss',
            startDate: this.startDate.format(this.format),
            endDate: this.endDate.format(this.format),
            // startDate: moment().subtract('days', 20),
            // endDate: moment().subtract('days', 1),
            opens: 'left'
          },
          function(start, end) {
            console.log('A date range was chosen: ' + start.format() + ' to ' + end.format());
            that.startDate = start;
            that.endDate = end;
            that.updateDownloadLink();
          }
        );

        var colors = d3.scale.category10();
        // basins
        // originally from http://www.great-lakes.net/gis/data/refdata.html
        console.log("Location: " + window.location.host);
        $.getJSON('http://' + window.location.host + '/assets/layers/glbasins_gen.geojson', function(data){
            console.log("adding basins");
            that.basins = L.geoJson(data,{
                style: function(feature) {
                    
                    switch (feature.properties.LAKEBASIN) {
                        case 'Erie': return {color: colors(0),"weight": 1, "opacity": 1};
                        case 'Huron': return {color: colors(1),"weight": 1, "opacity": 1};
                        case 'Michigan': return {color: colors(2),"weight": 1, "opacity": 1};
                        case 'Ontario': return {color: colors(3),"weight": 1, "opacity": 1};
                        case 'St. Lawrence': return {color: colors(4),"weight": 1, "opacity": 1};
                        case 'Superior': return {color: colors(5),"weight": 1, "opacity": 1};
                        default: return {color: colors(Math.round(Math.random()*10)),"weight": 1,};
                    }
                },
                onEachFeature: function (feature, layer) {
                    layer.bindPopup("<div class='search-popup'>"+feature.properties.LAKEBASIN+'</div>');
                },
            });
            // that.basins.addTo(that.map);
            that.mapControls.addOverlay(that.basins, "Basins");
            
        });

        // Initialise the FeatureGroup to store editable layers
        var drawnItems = new L.FeatureGroup();
        this.map.addLayer(drawnItems);

        var options = {
            position: 'topright',
            draw: {
                polyline: false,
                polygon: true,
                circle: true,
                rectangle: true,
                marker: false
            },
            edit: {
                featureGroup: drawnItems, //REQUIRED!!
                remove: true
            }
        };

        // Initialise the draw control and pass it the FeatureGroup of editable layers
        var drawControl = new L.Control.Draw(options);
        this.map.addControl(drawControl);

        var map = this.map; // FIXME


        this.sensorsfound.addTo(map);

        this.map.on('draw:created', function (e) {
            var type = e.layerType,
                layer = e.layer;

            if (type === 'polyline') {
                console.log("Bounds ");
                console.log(layer.getBounds());
                console.log("GeoJSON ");
                console.log(layer.toGeoJSON());
                console.log("LatLon " + layer.getLatLngs());
            }

            if (type === 'polygon') {
                console.log("GeoJSON ");
                console.log(layer.toGeoJSON());
                console.log("LatLon " + layer.getLatLngs());
            }


            if (type === 'rectangle') {
                console.log("Bounds ");
                console.log(layer.getBounds());
                console.log("GeoJSON ");
                console.log(layer.toGeoJSON());
                console.log("LatLon " + layer.getLatLngs());
            }

            if (type === 'circle') {
                console.log("LatLon " + layer.getLatLng());
                console.log("Radius " + layer.getRadius());
                that.eventBus.trigger("map:circle:select", {'latlng': layer.getLatLng(), 'radius': layer.getRadius()});
                drawnItems.clearLayers();
                that.sensorsfound.clearLayers();
                var latlng = layer.getLatLng();
                app.sensors.fetch({
                    dataType: "jsonp",
                    data: $.param({geocode: latlng.lat + ',' + latlng.lng + ',' + (layer.getRadius()/1000)}),
                    success: function (sensors) {

                        that.eventBus.trigger("sensors:selected", sensors.toJSON());

                        var geojsonMarkerOptions = {
                            radius: 2,
                            fillColor: "#009999",
                            color: "#009999",
                            weight: 5,
                            opacity: 1,
                            fillOpacity: 1
                        };

                        that.sensorsfound.addLayer(L.geoJson(sensors.toJSON(), {
                            onEachFeature: function (feature, layer) {
                                layer.bindPopup("<div class='search-popup'>"+feature.name+'</div>');
                            },
                            pointToLayer: function (feature, latlng) {
                                    return L.circleMarker(latlng, geojsonMarkerOptions);
                                }
                        }));
                    }
                });
            }

            if (type === 'marker') {
                console.log("LatLon " + layer.getLatLngs());
            }

            // Do whatever else you need to. (save to db, add to map etc)
            e.layer.addTo(drawnItems);
        });

        map.on('draw:edited', function () {
            // Update db to save latest changes.
        });

        map.on('draw:deleted', function () {
            that.sensorsfound.clearLayers();
            $('#download-json').addClass('disabled');
            $('#download-permalink').addClass('disabled');
        });


        this.stationService.getAll((function(self) {
            return function(data) {

                var geojsonMarkerOptions = {
                    radius: 2,
                    fillColor: "#ff7400",
                    color: "#ff7400",
                    weight: 5,
                    opacity: 1,
                    fillOpacity: 1
                };

                 var geoJsonLayer = L.geoJson(data, {
                    onEachFeature: function (feature, layer) {
                        layer.bindPopup("<div class='search-popup'>"+feature.name+'</div>');
                    },
                    pointToLayer: function (feature, latlng) {
                        return L.circleMarker(latlng, geojsonMarkerOptions);
                    }
                }).addTo(that.map);

                self.map.fitBounds(new L.featureGroup([geoJsonLayer]), {padding:[10,10]});
            };
        })(this));
	},


    updateDownloadLink : function(event) {
        var latlng = this.coordinates.latlng;
        var radius = this.coordinates.radius;
        var url = app.datapoints.url + '?' + $.param({geocode: latlng.lat + ',' + latlng.lng + ',' + (radius/1000), 
                    since: this.startDate.format(this.format), until: this.endDate.format(this.format)});
        $('#download-json').attr('href', url);
        $('#download-json').removeClass('disabled');
        $('#download-permalink').removeClass('disabled');
    },

    permalink : function(event) {
        var template = $("#permalink-template").html();
        $("#permalink").html(_.template(template,{url: $('#download-json').attr('href')}));
        $('#permalink-modal').modal('show');
    },
});
})(jQuery, Backbone, window);